//! # Compatibility layer.
//!
//! Provides compatibility when reading events generated by older versions of
//! Retis and in the Python library for accessing event fields that have been
//! modified. The two goals are in opposite directions:
//!
//! - Reading older events requires to make them look like the latest
//!   representation so they can be unmarshalled successfully.
//!
//! - Compatibility in the Python side requires to make the latest event look
//!   like older ones.

// Re-export compat.rs
#[allow(clippy::module_inception)]
pub(crate) mod compat;
pub(crate) use compat::*;

pub(crate) mod json;
